/*
 *  This sketch demonstrates how to set up a simple HTTP-like server.
 *  The server will set a GPIO pin depending on the request
 *    http://server_ip/gpio/0 will set the GPIO2 low,
 *    http://server_ip/gpio/1 will set the GPIO2 high
 *  server_ip is the IP address of the ESP8266 module, will be 
 *  printed to Serial when the module is connected.
 */

#include <ESP8266WiFi.h>
#include<Ticker.h>

#ifdef ESP8266
extern "C" {
    #include "user_interface.h"
}
#endif

#include <ESP8266HTTPClient.h>

// Set up output serial port (could be a SoftwareSerial
// if really wanted).
// Wired to the blue LED on an ESP-01 board, active LOW.
//
// Cannot be used simultaneously with Serial.print/println()
// calls, as TX is wired to the same pin. 
//
// UNLESS: You swap the TX pin using the alternate pinout.

Stream& ehConsolePort(Serial);

#define AP_MODE_MYSSID      "ESP_TESTAP"
#define AP_MODE_MYPW        "87654321"

#if 1
    #define STA_MODE_SHESSID    "xxxxxxx"
    #define STA_MODE_SHEPW      "xxxxxxx"
#else
    #define STA_MODE_SHESSID    AP_MODE_MYSSID
    #define STA_MODE_SHEPW      AP_MODE_MYPW
#endif

#define ACT_AS_AP 1
#define HTTP_SERVER 1
#define HTTP_CLIENT (!HTTP_SERVER)

#define ENABLE_DMP 1
#define IMU_STREAMOUT 0

// for http server
WiFiServer *server_p;

// for remote ip
char remote_ip[16] = {0};
String web_link;




/************************************************************
MPU9250_Basic
 Basic example sketch for MPU-9250 DMP Arduino Library 
Jim Lindblom @ SparkFun Electronics
original creation date: November 23, 2016
https://github.com/sparkfun/SparkFun_MPU9250_DMP_Arduino_Library

This example sketch demonstrates how to initialize the 
MPU-9250, and stream its sensor outputs to a serial monitor.

Development environment specifics:
Arduino IDE 1.6.12
SparkFun 9DoF Razor IMU M0

Supported Platforms:
- ATSAMD21 (Arduino Zero, SparkFun SAMD21 Breakouts)
*************************************************************/
#include "SparkFunMPU9250-DMP.h"


MPU9250_DMP imu;


void printIMUData(void)
{  
  // After calling update() the ax, ay, az, gx, gy, gz, mx,
  // my, mz, time, and/or temerature class variables are all
  // updated. Access them by placing the object. in front:

  // Use the calcAccel, calcGyro, and calcMag functions to
  // convert the raw sensor readings (signed 16-bit values)
  // to their respective units.
  float accelX = imu.calcAccel(imu.ax);
  float accelY = imu.calcAccel(imu.ay);
  float accelZ = imu.calcAccel(imu.az);
  float gyroX = imu.calcGyro(imu.gx);
  float gyroY = imu.calcGyro(imu.gy);
  float gyroZ = imu.calcGyro(imu.gz);
  float magX = imu.calcMag(imu.mx);
  float magY = imu.calcMag(imu.my);
  float magZ = imu.calcMag(imu.mz);

#if IMU_STREAMOUT

#if 1
    static uint8_t m[]={0x55, 0x55, 0, 0, 0, 0, 0, 0, (25&0xFF), ((25>>8)&0xFF), 0};
    
    unsigned short X=(unsigned short)(imu.calcAccel(imu.ax)*32768.f/16.f/9.8f);
    unsigned short Y=(unsigned short)(imu.calcAccel(imu.ay)*32768.f/16.f/9.8f);
    unsigned short Z=(unsigned short)(imu.calcAccel(imu.az)*32768.f/16.f/9.8f);
    m[2]=(X&0xFF);
    m[3]=((X>>8)&0xFF);
    m[4]=(Y&0xFF);
    m[5]=((Y>>8)&0xFF);
    m[6]=(Z&0xFF);
    m[7]=((Z>>8)&0xFF);
    m[1]=0x51; m[10]=0;
    for (int i=0; i<9; i++) m[10]+=m[i];
    m[10]&=0xFF;
    Serial.write(m, sizeof(m));
    
    X=(unsigned short)(imu.calcGyro(imu.gx)*32768.f/2000.f);
    Y=(unsigned short)(imu.calcGyro(imu.gy)*32768.f/2000.f);
    Z=(unsigned short)(imu.calcGyro(imu.gz)*32768.f/2000.f);
    m[2]=(X&0xFF);
    m[3]=((X>>8)&0xFF);
    m[4]=(Y&0xFF);
    m[5]=((Y>>8)&0xFF);
    m[6]=(Z&0xFF);
    m[7]=((Z>>8)&0xFF);
    m[1]=0x52; m[10]=0;
    for (int i=0; i<9; i++) m[10]+=m[i];
    m[10]&=0xFF;
    Serial.write(m, sizeof(m));
    
    X=(unsigned short)(imu.calcMag(imu.mx));
    Y=(unsigned short)(imu.calcMag(imu.my));
    Z=(unsigned short)(imu.calcMag(imu.mz));
    m[2]=(X&0xFF);
    m[3]=((X>>8)&0xFF);
    m[4]=(Y&0xFF);
    m[5]=((Y>>8)&0xFF);
    m[6]=(Z&0xFF);
    m[7]=((Z>>8)&0xFF);
    m[1]=0x54; m[10]=0;
    for (int i=0; i<9; i++) m[10]+=m[i];
    m[10]&=0xFF;
    Serial.write(m, sizeof(m));
#else
    int ax=(int)(imu.calcAccel(imu.ax)*100.f);
    int ay=(int)(imu.calcAccel(imu.ay)*100.f);
    int az=(int)(imu.calcAccel(imu.az)*100.f);
    Serial.write((unsigned char*)&ax, sizeof(int));
    Serial.write((unsigned char*)&ay, sizeof(int));
    Serial.write((unsigned char*)&az, sizeof(int));
    //imu.calcGyro(imu.gx)
    //imu.calcMag(imu.mx)
#endif

#else
  Serial.println("Accel: " + String(accelX) + ", " +
              String(accelY) + ", " + String(accelZ) + " g");
  Serial.println("Gyro: " + String(gyroX) + ", " +
              String(gyroY) + ", " + String(gyroZ) + " dps");
  Serial.println("Mag: " + String(magX) + ", " +
              String(magY) + ", " + String(magZ) + " uT");
  Serial.println("Time: " + String(imu.time) + " ms");
  Serial.println();
#endif
}

void setup_imu() 
{

  // Call imu.begin() to verify communication with and
  // initialize the MPU-9250 to it's default values.
  // Most functions return an error code - INV_SUCCESS (0)
  // indicates the IMU was present and successfully set up
  if (imu.begin() != INV_SUCCESS)
  {
    while (1)
    {
      Serial.println("Unable to communicate with MPU-9250");
      Serial.println("Check connections, and try again.");
      Serial.println();
      delay(5000);
    }
  }

  // Use setSensors to turn on or off MPU-9250 sensors.
  // Any of the following defines can be combined:
  // INV_XYZ_GYRO, INV_XYZ_ACCEL, INV_XYZ_COMPASS,
  // INV_X_GYRO, INV_Y_GYRO, or INV_Z_GYRO
  // Enable all sensors:
  imu.setSensors(INV_XYZ_GYRO | INV_XYZ_ACCEL | INV_XYZ_COMPASS);

  // Use setGyroFSR() and setAccelFSR() to configure the
  // gyroscope and accelerometer full scale ranges.
  // Gyro options are +/- 250, 500, 1000, or 2000 dps
  imu.setGyroFSR(2000); // Set gyro to 2000 dps
  // Accel options are +/- 2, 4, 8, or 16 g
  imu.setAccelFSR(2); // Set accel to +/-2g
  // Note: the MPU-9250's magnetometer FSR is set at 
  // +/- 4912 uT (micro-tesla's)

  // setLPF() can be used to set the digital low-pass filter
  // of the accelerometer and gyroscope.
  // Can be any of the following: 188, 98, 42, 20, 10, 5
  // (values are in Hz).
  imu.setLPF(5); // Set LPF corner frequency to 5Hz

  // The sample rate of the accel/gyro can be set using
  // setSampleRate. Acceptable values range from 4Hz to 1kHz
  imu.setSampleRate(10); // Set sample rate to 10Hz

  // Likewise, the compass (magnetometer) sample rate can be
  // set using the setCompassSampleRate() function.
  // This value can range between: 1-100Hz
  imu.setCompassSampleRate(10); // Set mag rate to 10Hz
}

void loop_imu() 
{
  // dataReady() checks to see if new accel/gyro data
  // is available. It will return a boolean true or false
  // (New magnetometer data cannot be checked, as the library
  //  runs that sensor in single-conversion mode.)
  if ( imu.dataReady() )
  {
    // Call update() to update the imu objects sensor data.
    // You can specify which sensors to update by combining
    // UPDATE_ACCEL, UPDATE_GYRO, UPDATE_COMPASS, and/or
    // UPDATE_TEMPERATURE.
    // (The update function defaults to accel, gyro, compass,
    //  so you don't have to specify these values.)
    imu.update(UPDATE_ACCEL | UPDATE_GYRO | UPDATE_COMPASS);
    printIMUData();

    delay(500);////kenwoo
  }
}


const char * const RST_REASONS[] =
{
    "REASON_DEFAULT_RST",
    "REASON_WDT_RST",
    "REASON_EXCEPTION_RST",
    "REASON_SOFT_WDT_RST",
    "REASON_SOFT_RESTART",
    "REASON_DEEP_SLEEP_AWAKE",
    "REASON_EXT_SYS_RST"
};

const char * const FLASH_SIZE_MAP_NAMES[] =
{
    "FLASH_SIZE_4M_MAP_256_256",
    "FLASH_SIZE_2M",
    "FLASH_SIZE_8M_MAP_512_512",
    "FLASH_SIZE_16M_MAP_512_512",
    "FLASH_SIZE_32M_MAP_512_512",
    "FLASH_SIZE_16M_MAP_1024_1024",
    "FLASH_SIZE_32M_MAP_1024_1024"
};

const char * const OP_MODE_NAMES[] 
{
    "NULL_MODE",
    "STATION_MODE",
    "SOFTAP_MODE",
    "STATIONAP_MODE"
};

const char * const AUTH_MODE_NAMES[] 
{
    "AUTH_OPEN",
    "AUTH_WEP",
    "AUTH_WPA_PSK",
    "AUTH_WPA2_PSK",
    "AUTH_WPA_WPA2_PSK",
    "AUTH_MAX"
};

const char * const PHY_MODE_NAMES[]
{
    "",
    "PHY_MODE_11B",
    "PHY_MODE_11G",
    "PHY_MODE_11N"
};

const char * const EVENT_NAMES[]
{
    "EVENT_STAMODE_CONNECTED",
    "EVENT_STAMODE_DISCONNECTED",
    "EVENT_STAMODE_AUTHMODE_CHANGE",
    "EVENT_STAMODE_GOT_IP",
    "EVENT_SOFTAPMODE_STACONNECTED",
    "EVENT_SOFTAPMODE_STADISCONNECTED",
    "EVENT_MAX"
};

const char * const EVENT_REASONS[]
{
    "",
    "REASON_UNSPECIFIED",
    "REASON_AUTH_EXPIRE",
    "REASON_AUTH_LEAVE",
    "REASON_ASSOC_EXPIRE",
    "REASON_ASSOC_TOOMANY",
    "REASON_NOT_AUTHED",
    "REASON_NOT_ASSOCED",
    "REASON_ASSOC_LEAVE",
    "REASON_ASSOC_NOT_AUTHED",
    "REASON_DISASSOC_PWRCAP_BAD",
    "REASON_DISASSOC_SUPCHAN_BAD",
    "REASON_IE_INVALID",
    "REASON_MIC_FAILURE",
    "REASON_4WAY_HANDSHAKE_TIMEOUT",
    "REASON_GROUP_KEY_UPDATE_TIMEOUT",
    "REASON_IE_IN_4WAY_DIFFERS",
    "REASON_GROUP_CIPHER_INVALID",
    "REASON_PAIRWISE_CIPHER_INVALID",
    "REASON_AKMP_INVALID",
    "REASON_UNSUPP_RSN_IE_VERSION",
    "REASON_INVALID_RSN_IE_CAP",
    "REASON_802_1X_AUTH_FAILED",
    "REASON_CIPHER_SUITE_REJECTED",
};

const char * const EVENT_REASONS_200[]
{
    "REASON_BEACON_TIMEOUT",
    "REASON_NO_AP_FOUND"
};

// The certificate is stored in PMEM
static const uint8_t x509[] PROGMEM = {
  0x30, 0x82, 0x01, 0x3d, 0x30, 0x81, 0xe8, 0x02, 0x09, 0x00, 0xfe, 0x56,
  0x46, 0xf2, 0x78, 0xc6, 0x51, 0x17, 0x30, 0x0d, 0x06, 0x09, 0x2a, 0x86,
  0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x0b, 0x05, 0x00, 0x30, 0x26, 0x31,
  0x10, 0x30, 0x0e, 0x06, 0x03, 0x55, 0x04, 0x0a, 0x0c, 0x07, 0x45, 0x53,
  0x50, 0x38, 0x32, 0x36, 0x36, 0x31, 0x12, 0x30, 0x10, 0x06, 0x03, 0x55,
  0x04, 0x03, 0x0c, 0x09, 0x31, 0x32, 0x37, 0x2e, 0x30, 0x2e, 0x30, 0x2e,
  0x31, 0x30, 0x1e, 0x17, 0x0d, 0x31, 0x37, 0x30, 0x33, 0x31, 0x38, 0x31,
  0x34, 0x34, 0x39, 0x31, 0x38, 0x5a, 0x17, 0x0d, 0x33, 0x30, 0x31, 0x31,
  0x32, 0x35, 0x31, 0x34, 0x34, 0x39, 0x31, 0x38, 0x5a, 0x30, 0x26, 0x31,
  0x10, 0x30, 0x0e, 0x06, 0x03, 0x55, 0x04, 0x0a, 0x0c, 0x07, 0x45, 0x53,
  0x50, 0x38, 0x32, 0x36, 0x36, 0x31, 0x12, 0x30, 0x10, 0x06, 0x03, 0x55,
  0x04, 0x03, 0x0c, 0x09, 0x31, 0x32, 0x37, 0x2e, 0x30, 0x2e, 0x30, 0x2e,
  0x31, 0x30, 0x5c, 0x30, 0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7,
  0x0d, 0x01, 0x01, 0x01, 0x05, 0x00, 0x03, 0x4b, 0x00, 0x30, 0x48, 0x02,
  0x41, 0x00, 0xc6, 0x72, 0x6c, 0x12, 0xe1, 0x20, 0x4d, 0x10, 0x0c, 0xf7,
  0x3a, 0x2a, 0x5a, 0x49, 0xe2, 0x2d, 0xc9, 0x7a, 0x63, 0x1d, 0xef, 0xc6,
  0xbb, 0xa3, 0xd6, 0x6f, 0x59, 0xcb, 0xd5, 0xf6, 0xbe, 0x34, 0x83, 0x33,
  0x50, 0x80, 0xec, 0x49, 0x63, 0xbf, 0xee, 0x59, 0x94, 0x67, 0x8b, 0x8d,
  0x81, 0x85, 0x23, 0x24, 0x06, 0x52, 0x76, 0x55, 0x9d, 0x18, 0x09, 0xb3,
  0x3c, 0x10, 0x40, 0x05, 0x01, 0xf3, 0x02, 0x03, 0x01, 0x00, 0x01, 0x30,
  0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x0b,
  0x05, 0x00, 0x03, 0x41, 0x00, 0x69, 0xdc, 0x6c, 0x9b, 0xa7, 0x62, 0x57,
  0x7e, 0x03, 0x01, 0x45, 0xad, 0x9a, 0x83, 0x90, 0x3a, 0xe7, 0xdf, 0xe8,
  0x8f, 0x46, 0x00, 0xd3, 0x5f, 0x2b, 0x0a, 0xde, 0x92, 0x1b, 0xc5, 0x04,
  0xc5, 0xc0, 0x76, 0xf4, 0xf6, 0x08, 0x36, 0x97, 0x27, 0x82, 0xf1, 0x60,
  0x76, 0xc2, 0xcd, 0x67, 0x6c, 0x4b, 0x6c, 0xca, 0xfd, 0x97, 0xfd, 0x33,
  0x9e, 0x12, 0x67, 0x6b, 0x98, 0x7e, 0xd5, 0x80, 0x8f
};

// And so is the key.  These could also be in DRAM
static const uint8_t rsakey[] PROGMEM = {
  0x30, 0x82, 0x01, 0x3a, 0x02, 0x01, 0x00, 0x02, 0x41, 0x00, 0xc6, 0x72,
  0x6c, 0x12, 0xe1, 0x20, 0x4d, 0x10, 0x0c, 0xf7, 0x3a, 0x2a, 0x5a, 0x49,
  0xe2, 0x2d, 0xc9, 0x7a, 0x63, 0x1d, 0xef, 0xc6, 0xbb, 0xa3, 0xd6, 0x6f,
  0x59, 0xcb, 0xd5, 0xf6, 0xbe, 0x34, 0x83, 0x33, 0x50, 0x80, 0xec, 0x49,
  0x63, 0xbf, 0xee, 0x59, 0x94, 0x67, 0x8b, 0x8d, 0x81, 0x85, 0x23, 0x24,
  0x06, 0x52, 0x76, 0x55, 0x9d, 0x18, 0x09, 0xb3, 0x3c, 0x10, 0x40, 0x05,
  0x01, 0xf3, 0x02, 0x03, 0x01, 0x00, 0x01, 0x02, 0x40, 0x35, 0x0b, 0x74,
  0xd3, 0xff, 0x15, 0x51, 0x44, 0x0f, 0x13, 0x2e, 0x9b, 0x0f, 0x93, 0x5c,
  0x3f, 0xfc, 0xf1, 0x17, 0xf9, 0x72, 0x94, 0x5e, 0xa7, 0xc6, 0xb3, 0xf0,
  0xfe, 0xc9, 0x6c, 0xb1, 0x1e, 0x83, 0xb3, 0xc6, 0x45, 0x3a, 0x25, 0x60,
  0x7c, 0x3d, 0x92, 0x7d, 0x53, 0xec, 0x49, 0x8d, 0xb5, 0x45, 0x10, 0x99,
  0x9b, 0xc6, 0x22, 0x3a, 0x68, 0xc7, 0x13, 0x4e, 0xb6, 0x04, 0x61, 0x21,
  0x01, 0x02, 0x21, 0x00, 0xea, 0x8c, 0x21, 0xd4, 0x7f, 0x3f, 0xb6, 0x91,
  0xfa, 0xf8, 0xb9, 0x2d, 0xcb, 0x36, 0x36, 0x02, 0x5f, 0xf0, 0x0c, 0x6e,
  0x87, 0xaa, 0x5c, 0x14, 0xf6, 0x56, 0x8e, 0x12, 0x92, 0x25, 0xde, 0xb3,
  0x02, 0x21, 0x00, 0xd8, 0x99, 0x01, 0xf1, 0x04, 0x0b, 0x98, 0xa3, 0x71,
  0x56, 0x1d, 0xea, 0x6f, 0x45, 0xd1, 0x36, 0x70, 0x76, 0x8b, 0xab, 0x69,
  0x30, 0x58, 0x9c, 0xe0, 0x45, 0x97, 0xe7, 0xb6, 0xb5, 0xef, 0xc1, 0x02,
  0x21, 0x00, 0xa2, 0x01, 0x06, 0xc0, 0xf2, 0xdf, 0xbc, 0x28, 0x1a, 0xb4,
  0xbf, 0x9b, 0x5c, 0xd8, 0x65, 0xf7, 0xbf, 0xf2, 0x5b, 0x73, 0xe0, 0xeb,
  0x0f, 0xcd, 0x3e, 0xd5, 0x4c, 0x2e, 0x91, 0x99, 0xec, 0xb7, 0x02, 0x20,
  0x4b, 0x9d, 0x46, 0xd7, 0x3c, 0x01, 0x4c, 0x5d, 0x2a, 0xb0, 0xd4, 0xaa,
  0xc6, 0x03, 0xca, 0xa0, 0xc5, 0xac, 0x2c, 0xe0, 0x3f, 0x4d, 0x98, 0x71,
  0xd3, 0xbd, 0x97, 0xe5, 0x55, 0x9c, 0xb8, 0x41, 0x02, 0x20, 0x02, 0x42,
  0x9f, 0xd1, 0x06, 0x35, 0x3b, 0x42, 0xf5, 0x64, 0xaf, 0x6d, 0xbf, 0xcd,
  0x2c, 0x3a, 0xcd, 0x0a, 0x9a, 0x4d, 0x7c, 0xad, 0x29, 0xd6, 0x36, 0x57,
  0xd5, 0xdf, 0x34, 0xeb, 0x26, 0x03
};

void wifi_event_handler_cb(System_Event_t * event)
{
    ehConsolePort.print(EVENT_NAMES[event->event]);
    ehConsolePort.print(" (");
    
    switch (event->event)
    {
        case EVENT_STAMODE_CONNECTED:
            break;
        case EVENT_STAMODE_DISCONNECTED:
            break;
        case EVENT_STAMODE_AUTHMODE_CHANGE:
            break;
        case EVENT_STAMODE_GOT_IP:
            break;
        case EVENT_SOFTAPMODE_STACONNECTED:
        case EVENT_SOFTAPMODE_STADISCONNECTED:
            {
                char mac[32] = {0};
                snprintf(mac, 32, MACSTR ", aid: %d" , MAC2STR(event->event_info.sta_connected.mac), event->event_info.sta_connected.aid);
                
                ehConsolePort.print(mac);
            }
            break;
    }

    ehConsolePort.println(")");
}

void print_softap_config(Stream & consolePort, softap_config const& config)
{
    consolePort.println();
    consolePort.println(F("SoftAP Configuration"));
    consolePort.println(F("--------------------"));

    consolePort.print(F("ssid:            "));
    consolePort.println((char *) config.ssid);

    consolePort.print(F("password:        "));
    consolePort.println((char *) config.password);

    consolePort.print(F("ssid_len:        "));
    consolePort.println(config.ssid_len);

    consolePort.print(F("channel:         "));
    consolePort.println(config.channel);

    consolePort.print(F("authmode:        "));
    consolePort.println(AUTH_MODE_NAMES[config.authmode]);

    consolePort.print(F("ssid_hidden:     "));
    consolePort.println(config.ssid_hidden);

    consolePort.print(F("max_connection:  "));
    consolePort.println(config.max_connection);

    consolePort.print(F("beacon_interval: "));
    consolePort.print(config.beacon_interval);
    consolePort.println("ms");

    consolePort.println(F("--------------------"));
    consolePort.println();
}

void print_system_info(Stream & consolePort)
{
    const rst_info * resetInfo = system_get_rst_info();
    consolePort.print(F("system_get_rst_info() reset reason: "));
    consolePort.println(RST_REASONS[resetInfo->reason]);

    consolePort.print(F("system_get_free_heap_size(): "));
    consolePort.println(system_get_free_heap_size());

    consolePort.print(F("system_get_os_print(): "));
    consolePort.println(system_get_os_print());
    system_set_os_print(1);
    consolePort.print(F("system_get_os_print(): "));
    consolePort.println(system_get_os_print());

    system_print_meminfo();

    consolePort.print(F("system_get_chip_id(): 0x"));
    consolePort.println(system_get_chip_id(), HEX);

    consolePort.print(F("system_get_sdk_version(): "));
    consolePort.println(system_get_sdk_version());

    consolePort.print(F("system_get_boot_version(): "));
    consolePort.println(system_get_boot_version());

    consolePort.print(F("system_get_userbin_addr(): 0x"));
    consolePort.println(system_get_userbin_addr(), HEX);

    consolePort.print(F("system_get_boot_mode(): "));
    consolePort.println(system_get_boot_mode() == 0 ? F("SYS_BOOT_ENHANCE_MODE") : F("SYS_BOOT_NORMAL_MODE"));

    consolePort.print(F("system_get_cpu_freq(): "));
    consolePort.println(system_get_cpu_freq());

    consolePort.print(F("spi_flash_get_id(): 0x"));
    consolePort.println(spi_flash_get_id(), HEX);

    consolePort.print(F("system_get_flash_size_map(): "));
    consolePort.println(FLASH_SIZE_MAP_NAMES[system_get_flash_size_map()]);
}

void print_wifi_general(Stream & consolePort)
{
    consolePort.print(F("wifi_get_channel(): "));
    consolePort.println(wifi_get_channel());
    
    consolePort.print(F("wifi_get_phy_mode(): "));
    consolePort.println(PHY_MODE_NAMES[wifi_get_phy_mode()]);
}

void secure_softap_config(softap_config * config, const char * ssid, const char * password)
{
    size_t ssidLen     = strlen(ssid)     < sizeof(config->ssid)     ? strlen(ssid)     : sizeof(config->ssid);
    size_t passwordLen = strlen(password) < sizeof(config->password) ? strlen(password) : sizeof(config->password);

    memset(config->ssid, 0, sizeof(config->ssid));
    memcpy(config->ssid, ssid, ssidLen);

    memset(config->password, 0, sizeof(config->password));
    memcpy(config->password, password, passwordLen);

    config->ssid_len = ssidLen;
    config->channel  = 1;
    config->authmode = AUTH_WPA2_PSK;
//    config->ssid_hidden = 1;
    config->max_connection = 4;
//    config->beacon_interval = 1000;
}

void enter_ap_mode(){

    WiFi.softAP(AP_MODE_MYSSID, AP_MODE_MYPW);///////////////
    IPAddress myIP = WiFi.softAPIP();///////////////
    Serial.print("AP IP ADDR: ");///////////////
    Serial.println(myIP);///////////////

    softap_config config;
    wifi_softap_get_config(&config);
    ///////////////////////secure_softap_config(&config, AP_MODE_MYSSID, AP_MODE_MYPW);
    ///////////////////////////////////////////////////wifi_softap_set_config(&config);
    print_softap_config(Serial, config);

    print_wifi_general(Serial);

    wifi_softap_dhcps_start();
    if (wifi_softap_dhcps_status()==DHCP_STARTED) Serial.println(F("DHCP started"));
    else Serial.println(F("DHCP stopped"));

    ip_info info;
    wifi_get_ip_info(SOFTAP_IF, &info);
    //Serial.printf("AP IP ADDR: %3d.%3d.%3d.%3d\r\n", ip4_addr1(&info.ip), ip4_addr2(&info.ip), ip4_addr3(&info.ip), ip4_addr4(&info.ip));
    //Serial.printf("AP GATEWAY: %3d.%3d.%3d.%3d\r\n", ip4_addr1(&info.gw), ip4_addr2(&info.gw), ip4_addr3(&info.gw), ip4_addr4(&info.gw));
    //Serial.printf("AP NETMASK: %3d.%3d.%3d.%3d\r\n", ip4_addr1(&info.netmask), ip4_addr2(&info.netmask), ip4_addr3(&info.netmask), ip4_addr4(&info.netmask));
    Serial.printf("AP IP ADDR: " IPSTR "\r\n", IP2STR(&info.ip));
    Serial.printf("AP GATEWAY: " IPSTR "\r\n", IP2STR(&info.gw));
    Serial.printf("AP NETMASK: " IPSTR "\r\n", IP2STR(&info.netmask));

    // This doesn't work on an ESP-01.
    // wifi_set_sleep_type(LIGHT_SLEEP_T);
    // Try this dirty little thing.
    // Doesn't work because ESP-01 module doesn't link XPD_DCDC -> RST.
    // ESP.deepSleep(15000);
}

void enter_station_mode(){
    unsigned char u=0;
    
    // Connect to WiFi network
    Serial.println();
    Serial.print("Connecting to ");
    Serial.println(STA_MODE_SHESSID);

    WiFi.mode(WIFI_STA);
    WiFi.begin(STA_MODE_SHESSID, STA_MODE_SHEPW);
    
    while (WiFi.status() != WL_CONNECTED) {
        delay(500);
        (u++%40)? Serial.print("."): Serial.print("\r\n.");
    }
    Serial.println("");
    Serial.println("WiFi connected");
    
    // Print the remote MAC address
    uint8_t macAddr[6];
    WiFi.macAddress(macAddr);
    Serial.printf("Remote MAC adr: %02X:%02X:%02X:%02X:%02X:%02X\r\n", macAddr[0], macAddr[1], macAddr[2], macAddr[3], macAddr[4], macAddr[5]);
    
    ip_info info;
    wifi_get_ip_info(STATION_IF, &info);
    //Serial.printf("IP ADDR: %3d.%3d.%3d.%3d\r\n", ip4_addr1(&info.ip), ip4_addr2(&info.ip), ip4_addr3(&info.ip), ip4_addr4(&info.ip));
    //Serial.printf("GATEWAY: %3d.%3d.%3d.%3d\r\n", ip4_addr1(&info.gw), ip4_addr2(&info.gw), ip4_addr3(&info.gw), ip4_addr4(&info.gw));
    //Serial.printf("NETMASK: %3d.%3d.%3d.%3d\r\n", ip4_addr1(&info.netmask), ip4_addr2(&info.netmask), ip4_addr3(&info.netmask), ip4_addr4(&info.netmask));
    Serial.printf("IP ADDR: " IPSTR "\r\n", IP2STR(&info.ip));
    Serial.printf("GATEWAY: " IPSTR "\r\n", IP2STR(&info.gw));
    Serial.printf("NETMASK: " IPSTR "\r\n", IP2STR(&info.netmask));

    // resolve the remote ip
    uint32 x=*(uint32*)(&info.ip)&*(uint32*)(&info.netmask);
    ip_addr y=*(ip_addr*)(&x);
    snprintf(remote_ip, 16, "%d.%d.%d.%d",\
        ip4_addr1(&y)? ip4_addr1(&y): 1,\
        ip4_addr2(&y)? ip4_addr2(&y): 1,\
        ip4_addr3(&y)? ip4_addr3(&y): 1,\
        ip4_addr4(&y)? ip4_addr4(&y): 1
        );
    Serial.print(F("Remote IP: "));
    Serial.println(remote_ip);

    Serial.printf("Default local hostname: %s\n", WiFi.hostname().c_str());
    Serial.print("DNS #1, #2 IP: ");
    WiFi.dnsIP().printTo(Serial);
    Serial.print(", ");
    WiFi.dnsIP(1).printTo(Serial);
    Serial.println();
    Serial.printf("RSSI: %d dBm\n", WiFi.RSSI());
    WiFi.printDiag(Serial);
    
    // Print the local IP address
    Serial.println("");
    Serial.print("Assigned the IP: ");
    Serial.println(WiFi.localIP());
}

int connected_station_status(){

    ///// this function for count couldn't sync the number of stat_infos retrieved so not to use.
    unsigned char number_client=wifi_softap_get_station_num(); // Count of stations

    struct station_info *stat_info;
    IPAddress address;
    int i;

    Serial.print("Total connected clients: ");
    Serial.println(number_client);

    for (i=0, stat_info=wifi_softap_get_station_info(); stat_info!=NULL && i<number_client; i++){
        address=(stat_info->ip).addr;

        snprintf(remote_ip, 16, "" IPSTR, IP2STR(&(stat_info->ip))); // used for outsider

        Serial.printf("Station %d. ", i+1);
        Serial.print("IP= ");
        Serial.print(address);
        Serial.print(" with MAC= ");
        Serial.print(stat_info->bssid[0],HEX);
        Serial.print(stat_info->bssid[1],HEX);
        Serial.print(stat_info->bssid[2],HEX);
        Serial.print(stat_info->bssid[3],HEX);
        Serial.print(stat_info->bssid[4],HEX);
        Serial.print(stat_info->bssid[5],HEX);
        Serial.println();

        stat_info = STAILQ_NEXT(stat_info, next);
    }
    wifi_softap_free_station_info();
    delay(100);
    return i; // not the number_client
}

WiFiServer* create_http_server(){
    // Create an instance of the server
    // specify the 80 port to listen on as an argument
    WiFiServer *a=new WiFiServer(80);

    ///WiFiServerSecure *a=new WiFiServerSecure(443);
    ///a->setServerKeyAndCert_P(rsakey, sizeof(rsakey), x509, sizeof(x509));
    ///a->begin();
    
    // Start the server
    a->begin();
    Serial.println("HTTP Server started");
    return a;
}

void http_client_connect(const char *url){

#if 0 // faster but often got refuse connection
    HTTPClient http;
    
    Serial.print("[HTTP] begin...\r\n");
    
    // configure traged server and url
    //http.begin("https://192.168.1.12/test.html", "7a 9c f4 db 40 d3 62 5a 6e 21 bc 5c cc 66 c8 3e a1 45 59 38"); //HTTPS
    http.begin(url); //HTTP
    Serial.print("[HTTP] GET...\r\n");
    // start connection and send HTTP header
    int httpCode = http.GET();

    // httpCode will be negative on error
    if (httpCode > 0){
        // HTTP header has been send and Server response header has been handled
        Serial.printf("[HTTP] GET... code: %d\n", httpCode);
        
        // file found at server
        if (httpCode == HTTP_CODE_OK){ // the connection may be broken and stuck here////////////////////////////////////////////////////
          Serial.println(http.getString());
        }
        else Serial.printf("[HTTP] GET... failed, error: %s\n", http.errorToString(httpCode).c_str());
    }
    else Serial.printf("[HTTP] Protocol failed, error: %s\n", http.errorToString(httpCode).c_str());
    
    http.end();
    delay(150);

#else // slower but always success

  // Use WiFiClient class to create TCP connections
  WiFiClient client;
  const int httpPort = 80;
  if (!client.connect(remote_ip, httpPort)) {
    Serial.println("");
    Serial.print("------------>");
    Serial.println("connection failed");
    Serial.println("");
    return;
  }

  Serial.print("Requesting URL: ");
  Serial.println(url);
  
  // This will send the request to the server
  client.print(String("GET ") + url + " HTTP/1.1\r\n" +
               "Host: " + remote_ip + "\r\n" + 
               "Connection: close\r\n\r\n");
  unsigned long timeout = millis();
  while (client.available() == 0) {
    if (millis() - timeout > 5000) {
      Serial.println(">>> Client Timeout !");
      client.stop();
      return;
    }
  }
  
  // Read all the lines of the reply from server and print them to Serial
  while(client.available()){
    String line = client.readStringUntil('\r');
    Serial.print(line);
  }

#endif

}

struct PWM2ms{
    int Duty;
    int Cycle;
    int Pin;

    PWM2ms(){memset(&(*this), 0, sizeof(*this));};
    void Driving();

protected:
    int PrevDuty;
    int PrevCycle;
    int PerMs;
    unsigned ReloadH;
    unsigned ReloadL;
    unsigned CounterH;
    unsigned CounterL;
    unsigned DelayH;
    unsigned DelayL;
};

// * set the GPIO pin as output before gening PWM
// * this function would be put in a per 1-ms revisit manner
// * cycle time: (H-level>1ms and L-level>1ms), 0<duty cycle<100 must be restricted elsewhere
void PWM2ms::Driving(){

    int t=micros();
    if (PerMs>t) return;

    if (Cycle!=PrevCycle){
        DelayH=Cycle*Duty*10;               // temporarily get the HIGH in microseconds
        ReloadH=DelayH/1000;                // the reload counts for HIGH
        DelayH%=1000;                       // the blocking delay in HIGH in microseconds(<1000us)
        DelayL=Cycle*(100-Duty)*10;         // temporarily get the LOW in microseconds
        ReloadL=DelayL/1000;                // the reload counts for LOW
        DelayL%=1000;                       // the blocking delay in LOW in microseconds(<1000us)
        PrevCycle=Cycle;
        PrevDuty=Duty;
        CounterH=CounterL=0;
    }
    if (Duty!=PrevDuty){
        DelayH=Cycle*Duty*10;               // temporarily get the HIGH in microseconds
        ReloadH=DelayH/1000;                // the reload counts for HIGH
        DelayH%=1000;                       // the blocking delay in HIGH in microseconds(<1000us)
        DelayL=Cycle*(100-Duty)*10;         // temporarily get the LOW in microseconds
        ReloadL=DelayL/1000;                // the reload counts for LOW
        DelayL%=1000;                       // the blocking delay in LOW in microseconds(<1000us)
        PrevDuty=Duty;
        CounterH=CounterL=0;
    }
    if (!CounterL){
        if (CounterH) CounterH--;
        else {
            ///// exception case 1, to prevent Duty==100 but 100%LOW condition. So Duty<100 is the sacrifice.
            /////CounterL=ReloadL;
            if (!ReloadL) CounterL=1;/////
            else CounterL=ReloadL;/////
            
            delayMicroseconds(DelayH);
            digitalWrite(Pin, LOW);
        }
    }
    else {
        CounterL--;
        if (!CounterL){
            CounterH=ReloadH;
            delayMicroseconds(DelayL);
            
            ///// exception case 2, while Duty==0 but got a pulse; no prevention.
            digitalWrite(Pin, HIGH);/////
        }
    }
    PerMs=t+1000;
}

// the example
void /*IRAM_ATTR*/ Revisit1ms(){
    extern PWM2ms pwm12, pwm13, pwm14, pwm15;
    static int cnt=0, sentinal=0, i=0;

    if (cnt>=100) sentinal=-1;
    else if (cnt<=0) sentinal=1;
    
    pwm12.Duty=cnt;
    pwm13.Duty=cnt;
    pwm14.Duty=cnt;
    pwm15.Duty=cnt;

    if (sentinal==1) pwm13.Cycle=15;
    else pwm13.Cycle=7;

    pwm12.Driving();
    pwm13.Driving();
    pwm14.Driving();
    pwm15.Driving();

    if (++i>200){
        i=0;
        cnt+=sentinal;
    }
}

static inline uint32_t MicrosecondsToCycles(uint32_t microseconds){
    return /*system_get_cpu_freq()*/80*microseconds;
}

static void initTimer(void(*Farg)()){
    ets_intr_lock();
    timer1_disable();
    timer1_isr_init();
    timer1_attachInterrupt(Farg);
    timer1_write(MicrosecondsToCycles(1000));
    timer1_enable(TIM_DIV1, TIM_EDGE, TIM_LOOP);
    ets_intr_unlock();
}

static void deinitTimer(){
    timer1_attachInterrupt(NULL);
    timer1_disable();
    timer1_isr_init();
}


/* ticker example
Ticker flipper;
flipper.attach(3, flip);
int count = 0;
void flip() {
    digitalWrite(LED_BUILTIN, !digitalRead(LED_BUILTIN));
    ++count;

    // when the counter reaches a certain value, start blinking like crazy
    if ((count % 20)==0){
        if ((count % 120)==0) flipper.detach();
        else flipper.attach(0.5, flip);
    }
}
*/


void pre_setup()
{
    // Reuse default Serial port rate, so the bootloader
    // messages are also readable.
    
    Serial.begin(112500);

    // Try pushing frequency to 160MHz.
    system_update_cpu_freq(SYS_CPU_160MHZ);

    Serial.println();
    Serial.println(F("ESP starting."));

    // System usually boots up in about 200ms.
    
    Serial.print(F("system_get_time(): "));
    Serial.println(system_get_time());

    // set_event_handler_cb_stream(Serial);
    /////////////////////wifi_set_event_handler_cb(wifi_event_handler_cb);

    print_system_info(Serial);

    Serial.print(F("wifi_get_opmode(): "));
    Serial.print(wifi_get_opmode());
    Serial.print(F(" - "));
    Serial.println(OP_MODE_NAMES[wifi_get_opmode()]);

    Serial.print(F("wifi_get_opmode_default(): "));
    Serial.print(wifi_get_opmode_default());
    Serial.print(F(" - "));
    Serial.println(OP_MODE_NAMES[wifi_get_opmode_default()]);

    Serial.print(F("wifi_get_broadcast_if(): "));
    Serial.println(wifi_get_broadcast_if());
}

void post_setup(){
    pinMode(LED_BUILTIN, OUTPUT); // Initialize the LED_BUILTIN pin (2) as an output
    digitalWrite(LED_BUILTIN, HIGH); // Turn the LED off

    Serial.printf("getResetReason: %s\r\n",         ESP.getResetReason().c_str());
    Serial.printf("getFreeHeap: %d\r\n",            ESP.getFreeHeap());
    Serial.printf("getChipId: %d\r\n",              ESP.getChipId());
    Serial.printf("getCoreVersion: %s\r\n",         ESP.getCoreVersion().c_str());
    Serial.printf("getSdkVersion: %s\r\n",          ESP.getSdkVersion());
    Serial.printf("getCpuFreqMHz: %u\r\n",          ESP.getCpuFreqMHz());
    Serial.printf("getSketchSize: %u\r\n",          ESP.getSketchSize());
    Serial.printf("getFreeSketchSpace: %u\r\n",     ESP.getFreeSketchSpace());
    Serial.printf("getSketchMD5: %s\r\n",           ESP.getSketchMD5().c_str());
    Serial.printf("getFlashChipId: %d\r\n",         ESP.getFlashChipId());
    Serial.printf("getFlashChipSize: %u\r\n",       ESP.getFlashChipSize());
    Serial.printf("getFlashChipRealSize: %u\r\n",   ESP.getFlashChipRealSize());
    Serial.printf("getFlashChipSpeed: %u\r\n",      ESP.getFlashChipSpeed());
    Serial.printf("getCycleCount: %u\r\n",          ESP.getCycleCount());
    Serial.printf("getVcc: %d\r\n",                 ESP.getVcc());
}



PWM2ms pwm12, pwm13, pwm14, pwm15;



void setup(){
    pre_setup();

#if ACT_AS_AP
    enter_ap_mode();
#else
    enter_station_mode();
#endif
    
    post_setup();

#if HTTP_SERVER
    // acted as an http server
    server_p=create_http_server();
#else // HTTP_CLIENT

    #if !ACT_AS_AP // the remote ip of http server is right valid
        // acted as an http client
        web_link="http://";
        web_link+=remote_ip;
        web_link+="/gpio/";
    #endif

#endif

    // PWM init
    pinMode(12, OUTPUT);
    pwm12.Pin=12;
    pinMode(13, OUTPUT);
    pwm13.Pin=13;
    pinMode(14, OUTPUT);
    pwm14.Pin=14;
    pinMode(15, OUTPUT);
    pwm15.Pin=15;
    pwm12.Cycle=pwm14.Cycle=pwm15.Cycle=100;
    initTimer(Revisit1ms);

#if ENABLE_DMP
    setup_imu();
    #if IMU_STREAMOUT
        delay(60000);
    #endif
#endif
}


void loop() {

#if ENABLE_DMP
    loop_imu();
#endif

#if HTTP_CLIENT

    #if ACT_AS_AP
        {
            static bool rip_ready=false;
            if (!rip_ready && connected_station_status()){
                rip_ready=true; // evaluate only once
                web_link="http://";
                web_link+=remote_ip;
                web_link+="/gpio/";
            }
            else if (!rip_ready){
                delay(5000);
                return;
            }
        }
    #endif

    static int x=0;
    http_client_connect((web_link+(x^=1)).c_str());
    delay(50);
    return;

#else // HTTP_SERVER

    WiFiServer &http_server=*server_p;
    
    // Check if a client has connected
    WiFiClient client = http_server.available();
    if (!client) return;
    
    // Wait until the client sends some data
    Serial.println("new client");
    
    
    // polling
    for (int x=0; !client.available(); x++){
        if (x>10){
            Serial.println("client missing");
            client.flush();
            client.stop();
            return;
        }
        delay(5);
    }
    
    
    // Read the first line of the request
    String req = client.readStringUntil('\r');
    Serial.println(req);
    
    // Match the request
    int val;
    if (req.indexOf("/gpio/0") != -1){
        val = 1;
    }
    else if (req.indexOf("/gpio/1") != -1){
        val = 0;
    }
    else {
        Serial.println("invalid request");
        client.print("HTTP/1.1 404 Not Found\r\nContent-Type: text/html\r\n\r\n<!DOCTYPE HTML>\r\n<html><body>Not found</body></html>");
        client.flush();
        client.stop();
        return;
    }
    
    
    // Set GPIO2 according to the request
    digitalWrite(2, val);
    
    // Prepare the response
    String s = "HTTP/1.1 200 OK\r\nContent-Type: text/html\r\n\r\n<!DOCTYPE HTML>\r\n<html>\r\nLED is now ";
    s += (val)? "OFF": "ON";

#if ENABLE_DMP
    // only be read once/////////////////////////////////////////////////////
    s+="\r\nAccel: [" + String(imu.calcAccel(imu.ax)) +\
                    ", " + String(imu.calcAccel(imu.ay)) + ", " + String(imu.calcAccel(imu.az)) + " g]\r\n";
    s+="Gyro:  [" + String(imu.calcGyro(imu.gx)) +\
                    ", " + String(imu.calcGyro(imu.gy)) + ", " + String(imu.calcGyro(imu.gz)) + " dps]\r\n";
    s+="Mag:   [" + String(imu.calcMag(imu.mx)) +\
                    ", " + String(imu.calcMag(imu.my)) + ", " + String(imu.calcMag(imu.mz)) + " uT]\r\n";
#endif

    s += "</html>\r\n";
    // Send the response to the client
    client.print(s);
    delay(1);
    Serial.println("Client disonnected");
    
    client.flush();
    client.stop();
    // The client will actually be disconnected
    // when the function returns and 'client' object is detroyed

#endif
}
